#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#       injector.py - Windows DLL Injector
#       
#       Copyright 2014 Recursos Python - www.recursospython.com
#       

"""
    Simple Windows DLL Injection Script / Module.
    Breaking the PEP 8 naming convention to achieve
    consistency with the Windows API convention.
"""


__all__ = ["InjectLibrary", "inject_library",
           "GetProcessIdByName", "get_process_id_by_name",
           "GetLastError", "get_last_error"]


# Access the Windows API via ctypess
from ctypes import c_ulong, c_char, pointer, sizeof, Structure, windll

# Necessary API functions
OpenProcess = windll.kernel32.OpenProcess
VirtualAllocEx = windll.kernel32.VirtualAllocEx
WriteProcessMemory = windll.kernel32.WriteProcessMemory
GetModuleHandle = windll.kernel32.GetModuleHandleA
GetProcAddress = windll.kernel32.GetProcAddress
CreateRemoteThread = windll.kernel32.CreateRemoteThread
CloseHandle = windll.kernel32.CloseHandle
CreateToolhelp32Snapshot = windll.kernel32.CreateToolhelp32Snapshot
Process32First = windll.kernel32.Process32First
Process32Next = windll.kernel32.Process32Next
_GetLastError = windll.kernel32.GetLastError

# Necessary Constants
PROCESS_CREATE_THREAD = 0x0002
PROCESS_QUERY_INFORMATION = 0x0400
PROCESS_VM_OPERATION = 0x0008
PROCESS_VM_WRITE = 0x0020
PROCESS_VM_READ = 0x0010
CREATE_THREAD_ACCESS = (PROCESS_CREATE_THREAD |
                        PROCESS_QUERY_INFORMATION |
                        PROCESS_VM_OPERATION |
                        PROCESS_VM_WRITE | PROCESS_VM_READ)

# VirtualAllocEx() constants
MEM_COMMIT = 0x00001000
MEM_RESERVE = 0x00002000
PAGE_READWRITE = 0x04

# CreateToolhelp32Snapshot() structures and constants
class PROCESSENTRY32(Structure):
     _fields_ = [("dwSize", c_ulong),
                 ("cntUsage", c_ulong),
                 ("th32ProcessID", c_ulong),
                 ("th32DefaultHeapID", c_ulong),
                 ("th32ModuleID", c_ulong),
                 ("cntThreads", c_ulong),
                 ("th32ParentProcessID", c_ulong),
                 ("pcPriClassBase", c_ulong),
                 ("dwFlags", c_ulong),
                 ("szExeFile", c_char * 260)]
TH32CS_SNAPPROCESS = 0x00000002


def GetProcessIdByName(sProcessName):
    """
    Retrieves the first matching process ID.
    If there's no matching process, the return value is None.
    """
    dwRetVal = None
    
    # An entry from a list of the processes
    pe32 = PROCESSENTRY32()
    pe32.dwSize = sizeof(PROCESSENTRY32)
    
    # Create snapshot
    hSnapshot = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, 0)
    if not hSnapshot > 0:
        raise RuntimeError, "Could not create snapshot."
    
    # Get first process
    if not Process32First(hSnapshot, pointer(pe32)):
        raise RuntimeError, "Could not get processes list."
    
    # Iterate into the processes list
    while Process32Next(hSnapshot, pointer(pe32)):
        if pe32.szExeFile == sProcessName:
            dwRetVal = pe32.th32ProcessID
            break
    
    # Release snapshot
    CloseHandle(hSnapshot)
    
    return dwRetVal

get_process_id_by_name = GetProcessIdByName


def InjectLibrary(sLibraryPath, dwProcessId, bOutput=False):
    """
    Injects the specified library at sLibraryPath into the
    process matching with the dwProcessId ID.
    """
    def _Output(s):
        if bOutput: print(s)
    
    # Get target process handle
    hProcess = OpenProcess(CREATE_THREAD_ACCESS, 0, dwProcessId)
    
    # Check if OpenProcess() succeed
    if not hProcess:
        error = "Could not open target process."
        _Output(error)
        raise RuntimeError, error
    
    dwLibraryPathLength = len(sLibraryPath)
    
    # Allocate enough memory for the library path string
    pRemoteString = VirtualAllocEx(hProcess, 0, dwLibraryPathLength,
                                   MEM_RESERVE | MEM_COMMIT,
                                   PAGE_READWRITE)
    
    # Check if VirtualAllocEx() succeed
    if not pRemoteString:
        error = "Could not allocate memory."
        _Output(error)
        raise MemoryError, error
    
    # Write the remote string into the target process
    if not WriteProcessMemory(hProcess, pRemoteString, sLibraryPath,
                              dwLibraryPathLength, None):
        error = "Could not write memory into target process."
        _Output(error)
        raise MemoryError, error
    
    # Load kernel32
    hKernel32 = GetModuleHandle("kernel32")
    if not hKernel32:
        error = "Could not load Kernel32.dll."
        _Output(error)
        raise RuntimeError, error
    
    # Get the LoadLibraryA() function
    pLoadLibrary = GetProcAddress(hKernel32, "LoadLibraryA")
    if not pLoadLibrary:
        error = "Could not load LoadLibraryA()."
        _Output(error)
        raise RuntimeError, error
    
    # Create and launch a remote thread with the DLL entrypoint
    hThread = CreateRemoteThread(hProcess, None, 0, pLoadLibrary,
                                 pRemoteString, 0, None)
    if not hThread:
        error = "Could not create remote thread."
        _Output(error)
        raise RuntimeError, error
    
    # If injected, close target process
    CloseHandle(hProcess)

inject_library = InjectLibrary


def GetLastError():
    """
    Retrieves the last error number.
    """
    return _GetLastError()

get_last_error = GetLastError


def main():
    from sys import argv
    
    # Check arguments
    try:
        sLibraryPath = argv[1]
        processData = argv[2]
    except IndexError:
        print("Invalid usage. Please use injector.py <dll_path> <proc_id_or_name>.")
    else:
        # Convert PID to a numeric value, if possible
        try:
            processData = int(processData)
        except ValueError:
            # Otherwise it's supposed to be the process name
            processData = GetProcessIdByName(processData)
        
        if processData is not None:
            # Inject into target process
            InjectLibrary(sLibraryPath, processData, True)
        else:
            print("Invalid process ID or name.")


if __name__ == "__main__":
    main()
